@@ -123,15 +123,28 @@ tpm_g = function(Z, beta, byrow = FALSE, ad = NULL, report = TRUE){
     # }
     
     expEta = exp(Z %*% t(beta))
-    Gamma = array(NaN, dim = c(N, N, nrow(expEta)))
+    Gamma = array(1, dim = c(N, N, nrow(expEta)))
+
+    ## Loop over entries (stuff over time happens vectorised which speeds up the tape)
+    col_ind <- 1
+    for(i in seq_len(N)){ # loop over rows
+      for(j in seq_len(N)){ # loop over columns
+        if(j != i){ # only if non-diagonal
+          if(byrow){
+            Gamma[i, j, ] <- expEta[, col_ind]
+          } else{
+            Gamma[j, i, ] <- expEta[, col_ind]
+          }
+          # increase col_ind by one
+          col_ind = col_ind + 1
+        }
+      }
+    }
     
-    for(t in 1:nrow(expEta)){
-      G = diag(N)
-      G[!G] = expEta[t,]
-      
-      if(byrow) G = t(G) # transpose if necessary
-      
-      Gamma[,,t] = G / rowSums(G)
+    # Normalise rows to sum to 1
+    for(i in seq_len(N)){
+      # transposing is necessary because Gamma[i,,] / colSums(Gamma[i,,]) does not work as expected
+      Gamma[i, , ] <- t(t(Gamma[i, , ]) / rowSums(t(Gamma[i, , ])))
     }
   }
   
